#!/bin/bash

# Contact Jangwon Kim for any question
#  jangwon@usc.edu

# A few notes
# - CD1: using the Contrastive-Divergence algorithm with one step of Markov-chain Monte Carlo sampling
# - This code does not perform feature transformation, but use the provided features as they are.

train_cmd="run.pl"
decode_cmd="run.pl"
source path.sh
export LC_ALL=C

###################################
# config for training DNN
###################################

# NN
hid_layers=1       # nr. of hidden layers (prior to sotfmax or bottleneck)
hid_dim=4          # select hidden dimension

# LABELS
num_tgt=2          # the number of output classes in the MLP (default is autodetect)

# TRAINING OPTIONS
learn_rate=0.008    # initial learning rate
train_opts=        # options, passed to the training script
train_tool=        # optionally change the training tool
frame_weights=     # per-frame weights for gradient weighting
momentum=0         
l1_penalty=0
l2_penalty=0
cross_val=true
# data processing
minibatch_size=256   # train on only one batch (256 samples) for parallel computing
randomizer_size=32768
randomizer_seed=777
# learn rate scheduling
max_iters=20
min_iters= 
#start_halving_inc=0.5
#end_halving_inc=0.1
start_halving_impr=0.01
end_halving_impr=0.001
halving_factor=0.5
# misc.
verbose=1
# tool
train_tool="nnet-train-frmshuff"
frame_weights=

# OTHER
use_gpu_id=-1      # manually select GPU id to run on, (-1 disables GPU)
seed=777    	   # seed value used for training data shuffling and initialization

# End configuration.



echo "$0 $@"  # Print the command line for logging

[ -f path.sh ] && . ./path.sh;

. parse_options.sh || exit 1;

if [ $# != 6 ]; then
   echo "Usage: bash $0 <data-train> <data-dev> <labels_tr> <labels_dev> <dbn> <exp-dir>"
   echo " e.g.: bash $0 feat/gauss_2D_2C/train feat/gauss_2D_2C/dev data/gauss_2D_2C/train/label data/gauss_2D_2C/dev/label exp/rbm_pretrain/3.dbn exp/nnet"
   echo "main options (for others, see top of script file)"
   echo "  --config <config-file>  # config containing options"
   exit 1;
fi

data_tr=$1
data_cv=$2
labels_tr=$3
labels_dev=$4
dbn=$5
dir=$6

echo
echo "# INFO"
echo "$0 : Training Neural Network"
printf "\t dir       : $dir \n"
printf "\t Train-set : $data_tr $labels_tr \n"
printf "\t CV-set    : $data_cv $labels_dev \n"

mkdir -p $dir/{log,nnet}


###################################
# prepare features
###################################
echo
echo "# PREPARING FEATURES"
# shuffle the list
echo "Preparing train/cv lists :"
cat $data_tr/feats.scp | utils/shuffle_list.pl --srand ${seed:-777} > $dir/train.scp
cp $data_cv/feats.scp $dir/cv.scp
# print the list sizes
wc -l $dir/train.scp $dir/cv.scp

#create a maximum 10k utt subset for global cmvn estimates
head -n 10000 $dir/train.scp > $dir/train.scp.10k



###################################
# prepare feature pipeline 
###################################

# read the features
#feats_tr="ark:copy-feats scp:$dir/train.scp ark:- |"
#feats_cv="ark:copy-feats scp:$dir/cv.scp ark:- |"
feats_tr="scp:$dir/train.scp"
feats_cv="scp:$dir/cv.scp"
lab_tr="ark:cat $labels_tr | ali-to-post ark,t:- ark:- |"
lab_cv="ark:cat $labels_dev | ali-to-post ark,t:- ark:- |"


# get feature dim
echo "Getting feature dim : "
num_fea=$(feat-to-dim --print-args=false "$feats_tr" -)
echo "Feature dim is : $num_fea"



###################################
# initialize the nnet
###################################

echo 
echo "# NN-INITIALIZATION"
# make network prototype
mlp_proto=$dir/nnet.proto
echo "Genrating network prototype $mlp_proto"
utils/nnet/make_nnet_proto.py $num_fea $num_tgt $hid_layers $hid_dim >$mlp_proto || exit 1
# initialize
mlp_init=$dir/nnet.init; log=$dir/log/nnet_initialize.log
echo "Initializing $mlp_proto -> $mlp_init"
nnet-initialize $mlp_proto $mlp_init 2>$log || { cat $log; exit 1; }



###################################
# train the nnet
###################################

echo
echo "# START NN-TRAINING"

# choose mlp to start with
mlp_best=$mlp_init
mlp_base=${mlp_init##*/}; mlp_base=${mlp_base%.*}

# cross-validation on original network
log=$dir/log/iter00.initial.log; hostname>$log
$train_tool --cross-validate=$cross_val \
 --minibatch-size=$minibatch_size --randomizer-size=$randomizer_size --verbose=$verbose \
 ${frame_weights:+ "--frame-weights=$frame_weights"} \
 "$feats_cv" "$lab_cv" $mlp_best \
 2>> $log || exit 1;

loss=$(cat $dir/log/iter00.initial.log | grep "AvgLoss:" | tail -n 1 | awk '{ print $4; }')
loss_type=$(cat $dir/log/iter00.initial.log | grep "AvgLoss:" | tail -n 1 | awk '{ print $5; }')
echo "CROSSVAL PRERUN AVG.LOSS $(printf "%.4f" $loss) $loss_type"

# resume lr-halving (learning rate halving)
halving=0
# training
for iter in $(seq -w $max_iters); do
  echo -n "ITERATION $iter: "
  mlp_next=$dir/nnet/${mlp_base}_iter${iter}

  # training
  log=$dir/log/iter${iter}.tr.log; hostname>$log
  $train_tool \
   --learn-rate=$learn_rate --momentum=$momentum --l1-penalty=$l1_penalty --l2-penalty=$l2_penalty \
   --minibatch-size=$minibatch_size --randomizer-size=$randomizer_size --randomize=true --verbose=$verbose \
   --binary=true \
   ${frame_weights:+ "--frame-weights=$frame_weights"} \
   ${randomizer_seed:+ --randomizer-seed=$randomizer_seed} \
   "$feats_tr" "$lab_tr" $mlp_best $mlp_next \
   2>> $log || exit 1;

  tr_loss=$(cat $dir/log/iter${iter}.tr.log | grep "AvgLoss:" | tail -n 1 | awk '{ print $4; }')
  echo -n "TRAIN AVG.LOSS $(printf "%.4f" $tr_loss), (lrate$(printf "%.6g" $learn_rate)), "

  # cross-validation
  log=$dir/log/iter${iter}.cv.log; hostname>$log
  $train_tool --cross-validate=true \
   --minibatch-size=$minibatch_size --randomizer-size=$randomizer_size --verbose=$verbose \
   ${frame_weights:+ "--frame-weights=$frame_weights"} \
   "$feats_cv" "$lab_cv" $mlp_next \
   2>>$log || exit 1;

  loss_new=$(cat $dir/log/iter${iter}.cv.log | grep "AvgLoss:" | tail -n 1 | awk '{ print $4; }')
  echo -n "CROSSVAL AVG.LOSS $(printf "%.4f" $loss_new), "

  # accept or reject new parameters (based on objective function)
  loss_prev=$loss
  if [ "1" == "$(awk "BEGIN{print($loss_new<$loss);}")" ]; then
    loss=$loss_new
    mlp_best=$dir/nnet/${mlp_base}_iter${iter}_learnrate${learn_rate}_tr$(printf "%.4f" $tr_loss)_cv$(printf "%.4f" $loss_new)
    mv $mlp_next $mlp_best
    echo "nnet accepted ($(basename $mlp_best))"
    echo $mlp_best > $dir/.mlp_best
  else
    mlp_reject=$dir/nnet/${mlp_base}_iter${iter}_learnrate${learn_rate}_tr$(printf "%.4f" $tr_loss)_cv$(printf "%.4f" $loss_new)_rejected
    mv $mlp_next $mlp_reject
    echo "nnet rejected ($(basename $mlp_reject))"
  fi

  # create .done file as a mark that iteration is over
  touch $dir/.done_iter$iter

  # stopping criterion
  if [[ "1" == "$halving" && "1" == "$(awk "BEGIN{print(($loss_prev-$loss)/$loss_prev < $end_halving_impr)}")" ]]; then
    if [[ "$min_iters" != "" ]]; then
      if [ $min_iters -gt $iter ]; then
        echo we were supposed to finish, but we continue, min_iters : $min_iters
        continue
      fi
    fi
    echo finished, too small rel. improvement $(awk "BEGIN{print(($loss_prev-$loss)/$loss_prev)}")
    break
  fi

  # start annealing when improvement is low
  if [ "1" == "$(awk "BEGIN{print(($loss_prev-$loss)/$loss_prev < $start_halving_impr)}")" ]; then
    halving=1
    echo $halving >$dir/.halving
  fi

  # do annealing
  if [ "1" == "$halving" ]; then
    learn_rate=$(awk "BEGIN{print($learn_rate*$halving_factor)}")
    echo $learn_rate >$dir/.learn_rate
  fi
done

# select the best network
if [ $mlp_best != $mlp_init ]; then
  mlp_final=${mlp_best}_final_
  ( cd $dir/nnet; ln -s $(basename $mlp_best) $(basename $mlp_final); )
  ( cd $dir; ln -s nnet/$(basename $mlp_final) final.nnet; )
  echo "Succeeded training the Neural Network : $dir/final.nnet"
else
  "Error training neural network..."
  exit 1
fi



echo "$0 successfuly finished.. $dir"

sleep 3
exit 0

